import argparse
import os
import stat
import subprocess
import sys
import multiprocessing
from termcolor import colored
from reseasdk.helpers import render, info, notice, error, generating, \
    load_yaml, dict_to_strdict
from reseasdk.validators import \
    validate_packages_yml, validate_package_yml, \
    validate_build_yml, validate_global_yml


SHORT_HELP = "build an executable"
LONG_HELP = """
Usage: reseasdk build
"""

START_C_TEMPLATE = """\
#include <resea.h>
#undef PACKAGE_NAME
#define PACKAGE_NAME "start_apps"
#define DEFAULT_THREAD_SIZE 0x4000

Result core_create_thread(Id group, const UChar *name, Size name_size,
                          Id *r_thread, Id *r_group);
Result core_run_thread(Id thread, Addr entry, Addr arg,
                       Addr stack, Size stack_size);
void core_thread_start_threading(void);

{% for app in builtin_apps %}
void {{ app }}_startup();
{% endfor %}

void {{ current_package }}_test();

static void app_entrypoint (void (*startup)()){

  startup();
  BUG("startup() returned");
  for(;;); // FIXME
}


UNUSED static void test_entrypoint (void (*test)()){

  logprintf("[start_apps] TEST: start");
  test();
  logprintf("[start_apps] TEST: end");
  for(;;); // FIXME
}

void start_apps(void){
{% if with_thread %}
    Id thread, group;
{% endif %}

{% for app in builtin_apps %}
{%  if app != 'core' %}
{% if with_thread %}
    INFO("starting '{{ app }}'");
    core_create_thread(1, (const UChar *) "{{ app }}",
                       {{ app | length }}, &thread, &group);
    core_run_thread(thread, 
                    (Addr) app_entrypoint, (Addr) {{ app }}_startup,
                    (Addr) allocMemory(DEFAULT_THREAD_SIZE, ALLOCMEM_NORMAL),
                    DEFAULT_THREAD_SIZE);
{% else %}
    INFO("starting {{ app }} (direct-startup)");
    {{ app }}_startup();
{% endif %}
{% endif %}
{% endfor %}

{% if defaults['TEST'] == 'yes' %}
{% if with_thread %}
    INFO("starting 'test_{{ current_package }}'");
    core_create_thread(1, (const UChar *) "{{ app }}",
                       {{ app | length }}, &thread, &group);
    core_run_thread(thread, 
                    (Addr) app_entrypoint, (Addr) {{ app }}_startup,
                    (Addr) allocMemory(DEFAULT_THREAD_SIZE, ALLOCMEM_NORMAL),
                    DEFAULT_THREAD_SIZE);
{% else %}
    logprintf("[start_apps] TEST: start");
    {{ current_package }}_test();
    logprintf("[start_apps] TEST: <fail> threading is not enabled"
              "-- cannot start a test thread");
    logprintf("[start_apps] TEST: end");
{% endif %}
{% endif %}

{% if with_thread %}
    INFO("starting threading");
    core_thread_start_threading();
{% endif %}
}
"""

MAKEFILE_TEMPLATE = """\
.PHONY: _default
_default: default
# keep blank not to delete intermediate file (especially stub files)
.SECONDARY:
$(VERBOSE).SILENT:

# default build config
{% for k,v in defaults.items() %}
export {{ k }} = {{ v }}
{% endfor %}

# {{ filetype }}
default: $(BUILD_DIR)/{{ filetype }}
$(BUILD_DIR)/{{ filetype }}: \\
    {% for obj in objs %}
    {{ obj }} \\
    {% endfor %}

{% if filetype == 'application' %}
\t$(CMDECHO) LINK $@
\t$(HAL_LINK) $@ $^
{% else %}
\t$(CMDECHO) LD_R $@
\t$(LD_R) $@ $^
{% endif %}


{% for abbrev,ext,with_stub in langs %}
#  *.{{ ext }}
{% if with_stub %}
STUBS_{{ ext }} = \\
{% for stub in stubs %}
  $(BUILD_DIR)/stubs/{{ ext }}/$(STUB_PREFIX_{{ ext }}){{ stub }}\
$(STUB_SUFFIX_{{ ext }}) \\
{% endfor %}

$(BUILD_DIR)/stubs/{{ ext }}/$(STUB_PREFIX_{{ ext }})%\
$(STUB_SUFFIX_{{ ext }}): \
packages/%/package.yml
\t$(MKDIR) -p $(@D)
\t$(CMDECHO) GENSTUB $@
\tPACKAGE_NAME=$(PACKAGE_NAME) $(GENSTUB_{{ ext }}) $@ $<
{% endif %}
$(BUILD_DIR)/%.o: packages/%.{{ ext }} $(STUBS_{{ ext }}) $(BUILD_DIR)/Makefile
\t$(MKDIR) -p $(@D)
\t$(CMDECHO) '{{ abbrev }}' $@
\tPACKAGE_NAME=$(PACKAGE_NAME) $(COMPILE_{{ ext }}) $@ $<
{% endfor %}

# start.c
$(BUILD_DIR)/start.o: $(BUILD_DIR)/start.c $(STUBS_c) $(BUILD_DIR)/Makefile
\t$(MKDIR) -p $(@D)
\t$(CMDECHO) CC $@
\t$(COMPILE_c) $@ $<

{% for file in files %}
{% for required_by in file['required_by'] %}
{{ required_by }}: {{ file['path'] }}
{% endfor %}

{{ file['path'] }}: \\
{% for d in file['depends'] %}
    {{ d }}
{% endfor %}
\t$(MKDIR) -p $(@D)
\t$(CMDECHO) 'GEN' $@
{% for command in file['commands'] %}
\t{{ command }}
{% endfor %}
{% endfor %}

{% for package_name,config in configs.items() %}
# {{ package_name }}
$(BUILD_DIR)/{{ package_name }}/%: PACKAGE_NAME = {{ package_name }}
{% for k,v in config['set'].items() %}
$(BUILD_DIR)/{{ package_name }}/%: {{ k }} = {{ v }}
{% endfor %}
{% for k,v in config['append'].items() %}
$(BUILD_DIR)/{{ package_name }}/%: {{ k }} += {{ v }}
{% endfor %}
{% endfor %}
"""


def get_package_dir(package):
    cwd = os.getcwd()
    # look for packages.yml
    while True:
        os.chdir("..")
        if os.getcwd() == "/":
            error("packages.yml not found")
        if os.path.exists("packages.yml"):
            package_dir = os.getcwd()
            break

    yml = load_yaml('packages.yml', validator=validate_packages_yml)
    os.chdir(cwd)

    # get the path to the package
    try:
        path = os.path.join(package_dir, yml["packages"][package]["path"])
    except KeyError:
        error("package not found in packages.yml: '{}'".format(package))

    return path


def get_package(package):
    """ Prepare a package in 'packages' directory """

    if not os.path.exists('packages/' + package):
        os.makedirs('packages', exist_ok=True)
        os.chdir('packages')
        os.symlink(get_package_dir(package), package)
        os.chdir('..')


global_ymls = {}


def load_global_yml(package):
    if package in global_ymls:
        return global_ymls[package]

    get_package(package)
    try:
        yml = load_yaml('packages/{}/global.yml'.format(package),
                        validator=validate_global_yml)
    except FileNotFoundError:
        yml = {}
    global_ymls[package] = yml
    return yml


build_ymls = {}


def load_build_yml(package):
    if package in build_ymls:
        return build_ymls[package]

    get_package(package)
    try:
        yml = load_yaml('packages/{}/build.yml'.format(package),
                        validator=validate_build_yml)
    except FileNotFoundError:
        yml = {}
    build_ymls[package] = yml
    return yml


package_ymls = {}


def load_package_yml(package):
    if package in package_ymls:
        return package_ymls[package]

    get_package(package)
    yml = load_yaml('packages/{}/package.yml'.format(package),
                    validator=validate_package_yml)
    package_ymls[package] = yml
    return yml


def resolve_package_dependency(_packages, all_in_one):
    """ Returns list of required packages. """
    packages = []
    follows = ['library']
    if all_in_one:
        follows.append('application')
    for package in _packages:
        yml = load_package_yml(package)
        packages += yml['implements']
        packages += yml['uses']
        for require in yml['requires']:
            if load_package_yml(require)['category'] in follows:
                packages += resolve_package_dependency([require], all_in_one)
            packages.append(require)
        packages.append(package)
    return list(set(packages))


def filter_package_by_category(categories, packages):
    matched = []
    for package in packages:
        if load_package_yml(package)['category'] in categories:
            matched.append(package)
    return matched


def clear_current_line():
    """Clears the current line in the terminal."""
    print('\r\x1b[0K', end='')


def prettify_make_output(p):
    prev_is_cmd = False
    while True:
        l = p.stdout.readline().decode('utf-8').rstrip()
        if prev_is_cmd:
            clear_current_line()

        if l == '':
            break
        elif l.startswith('--> '):
            try:
                cmd, rest = l.lstrip('--> ').split(' ', 1)
                s = '{:<8} {}'.format(
                    colored(cmd, 'magenta', attrs=['bold']),
                    colored(rest, 'yellow'))
                print(s, end='')
                sys.stdout.flush()
                prev_is_cmd = True
            except ValueError:
                print(l)
                prev_is_cmd = False
        else:
            print(l)
            prev_is_cmd = False


def load_configs(args, defaults):
    defaults.setdefault('LD_R', '$(LD) -r -o')
    defaults.setdefault('MKDIR', 'mkdir')
    defaults.setdefault('CMDECHO', 'echo "-->"')
    defaults.setdefault('BUILTIN_APPS', '')

    if 'HAL' not in defaults:
        error('HAL is not speicified')  # TODO
    hal = defaults['HAL']
    build_dir = defaults['BUILD_DIR']
    current_package = defaults['PACKAGE']  # a package to be built
    # applicaitons to be embedded
    builtin_apps = defaults['BUILTIN_APPS'].split()
    packages = [current_package, hal] + builtin_apps  # packages to be embeded

    # generate the build directory
    if not os.path.exists(build_dir):
        os.makedirs(build_dir, exist_ok=True)

    if load_package_yml(current_package)['category'] == 'application':
        # current_package will be embedded
        builtin_apps.append(current_package)

    if load_package_yml(current_package)['category'] == 'interface':
        # FIXME: raise an exception instead
        error("the '{}' is an interface package".format(defaults['PACKAGE']))

    # test
    if defaults.get('TEST') == 'yes':
        if not load_build_yml(current_package).get('testable', False):
            # FIXME: raise an exception instead
            error("the '{}' is not testable".format(current_package))

        # an executable file will be built even if it is 'library'
        filetype = 'application'
        objs = []
        for obj in load_build_yml(current_package).get('test_objs', []):
            objs.append(build_dir + '/' + current_package + '/' + obj)
    elif len(builtin_apps) == 0:
        # an object file (.o) will be built (as a library)
        filetype = 'library'
        objs = []
    else:
        # an executable file will be built
        filetype = 'application'
        objs = []

    # load packages
    packages = resolve_package_dependency(
        packages, all_in_one=args['all_in_one'])

    configs = {}
    langs = []
    stubs = []
    files = []
    for package in packages:
        stubs.append(package)
        # FIXME: build config
        configs[package] = {'set': {}, 'append': {}}
        for k, v in load_build_yml(package).items():
            if k not in ['objs', 'test_objs']:  # exclude those
                if k.startswith("+"):
                    if k.lstrip("+") not in configs[package]['append']:
                        configs[package]['append'][k.lstrip("+")] = v.strip()
                    else:
                        configs[package]['append'][
                            k.lstrip("+")] += ' ' + v.strip()
                else:
                    configs[package]['set'][k] = v

        for k, v in load_global_yml(package).items():
            if k.startswith("+"):
                if k.lstrip("+") not in defaults:
                    defaults[k.lstrip("+")] = v.strip()
                else:
                    defaults[k.lstrip("+")] += ' ' + v.strip()
            else:
                if k not in defaults:
                    defaults[k] = v

    for package in filter_package_by_category(['application'], packages):
        if package != current_package:
            builtin_apps.append(package)

    for package in filter_package_by_category(
            ['application', 'library'], packages):
        for obj in load_build_yml(package).get('objs', []):
            objs.append(build_dir + '/' + package + '/' + obj)

    for package in filter_package_by_category(['library'], packages):
        yml = load_package_yml(package)
        implements = yml['implements']
        if "lang" in implements:
            for lang in yml["langs"]:
                langs.append((lang['abbrev'], lang['ext'], lang['stub']))
        if "hal" in implements:
            pass

    return locals()


def generate_files(args, cs):
    """ Generate files required to build """

    if cs['filetype'] == 'application':
        cs['with_thread'] = not args['single_app']

        # generate start.c
        with open(cs['build_dir'] + '/start.c', 'w') as f:
            f.write(render(START_C_TEMPLATE, cs))

        # c_lang is required because it is essential to compile start.c
        cs['packages'].append('c_lang')
        cs['objs'] += [cs['build_dir'] + "/start.o"]

    # generate Makefile
    makefile = render(MAKEFILE_TEMPLATE, cs)
    with open(cs['build_dir'] + '/Makefile', 'w') as f:
        f.write(makefile)

    if args["dump"]:
        print(makefile)

    return cs


def build(args):
    """Builds an executable."""

    defaults = {
        'BUILD_DIR': 'build/' + args['target'],
        'MAKEFLAGS': '-j' + str(multiprocessing.cpu_count())
    }
    for default in args['defaults']:
        if '=' not in default:
            error('invalid default variable (should be'
                  "FOO=bar form): '{}'".format(default))
        k, v = default.split('=', 1)
        defaults[k] = v

    if not os.path.exists('package.yml'):
        error("'package.yml' not found (are you in a package directory?)")

    if not defaults.get('PACKAGE'):
        yml = load_yaml('package.yml', validator=validate_package_yml)
        defaults['PACKAGE'] = yml['name']

    makefile = defaults['BUILD_DIR'] + '/Makefile'
    cs = load_configs(args, defaults)
    if args['r'] or not os.path.exists(makefile):
        generate_files(args, cs)

    # execute make(1)
    env = os.environ.copy()
    env.update(dict_to_strdict(cs['defaults']))
    try:
        p = subprocess.Popen(['make', '-f', makefile],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT,
                             env=env)
    except Exception as e:
        error('failed to execute make: ' + str(e))

    if args['no_prettify']:
        while True:
            l = p.stdout.readline().decode('utf-8').rstrip()
            if l == '':
                break
            print(l)
    else:
        prettify_make_output(p)

    if p.wait() != 0:
        error('error occurred during make')

    return cs


def main(args_):
    parser = argparse.ArgumentParser(prog='reseasdk build',
                                     description='build an executable')
    parser.add_argument('-r', action='store_true', help='regenerate Makefile')
    parser.add_argument('--target', default='release', help='the build target')
    parser.add_argument('--dump', action='store_true',
                        help='dump generated files for debugging')
    parser.add_argument('--no-prettify', action='store_true',
                        help="don't prettify output")
    parser.add_argument('--all-in-one', action='store_true',
                        help='embed all required applications so that'
                             'discovery() always succeed')
    parser.add_argument('--single-app', action='store_true',
                        help='no threading')
    parser.add_argument('defaults', nargs='*',
                        help='default variables (FOO=bar)')

    build(vars(parser.parse_args(args_)))



"""

Useful functions

"""
import os
import shlex
import subprocess
import sys
import yaml
import jinja2
from termcolor import colored, cprint


def error(msg):
    """Prints an error message and terminate the program."""
    sys.exit(colored('Resea SDK: {}'.format(msg), 'red'))


def notice(msg):
    """Prints an notice."""
    cprint('Resea SDK: {}'.format(msg), 'yellow')


def info(msg):
    """Prints an informational message."""
    cprint(msg, 'blue')


def _generating(cmd, target):
    """Returns log message for generating something."""

    return '  {}{}{}'.format(colored(cmd, 'magenta'),
                             (' ' * (16 - len(cmd))),
                             target)


def generating(cmd, target):
    """Prints a 'GEN somthing' message. """
    print(_generating(cmd, target))


def exec_cmd(cmd, ignore_failure=False, cwd=None):
    """Executes a command and returns output.

    cmd can be a list (argv) or str. If cmd is str, it will be
    converted by shltex.split().

    If ignore_failure is True, it returns a empty string
    when an error occurred during executing a command instead
    of raising an exception.

    If cwd is not None, it executes the command in the directory.

    >>> exec_cmd('echo hello!')
    'hello!\\n'
    >>> exec_cmd('./bin/sh -c "pwd"', cwd="/")
    '/\\n'
    >>> exec_cmd('exit 1', ignore_failure=True)
    ''
    >>> exec_cmd('this_command_does_not_exist')
    Traceback (most recent call last):
     ...
    FileNotFoundError: [Errno 2] No such file or directory: \
'this_command_does_not_exist'
    """

    if isinstance(cmd, str):
        cmd = shlex.split(cmd)

    if cwd is not None:
        old_cwd = os.getcwd()
        os.chdir(cwd)

    try:
        output = subprocess.check_output(cmd)
    except Exception as e:
        if ignore_failure:
            return ''
        else:
            raise e
    else:
        return output.decode('utf-8')
    finally:
        if cwd is not None:
            os.chdir(old_cwd)


def import_module(module):
    """Returns a module.

    For example, if you do import_module('foo.bar.baz'), it returns
    the 'baz' module instead of 'foo'.
    """
    m = __import__(module)
    for x in module.split('.'):
        m = getattr(m, x)
    return m


def render(tmpl, vars):
    """Renders a template by jinja2."""
    return jinja2.Environment(
        trim_blocks=True,
        lstrip_blocks=True).from_string(tmpl).render(vars)


def load_yaml(path, validator=None):
    """Loads a yaml file."""
    yml = yaml.safe_load(open(path))
    if validator is not None:
        validator(yml)
    return yml


def dict_to_strdict(d):
    for k, v in d.items():
        if isinstance(v, (list, tuple)):
            d[k] = " ".join(v)
        if isinstance(v, bool):
            d[k] = repr(v)
        elif isinstance(v, dict):
            d[k] = "dict_to_strdict: error: '{}' is dict".format(k)  # FIXME
    return d



import argparse
import hashlib
import os
import sys
import subprocess
import time
import threading
import watchdog
import watchdog.events
import watchdog.observers
from reseasdk.run import run_emulator
from reseasdk.helpers import info, notice, error, \
    dict_to_strdict
from reseasdk.commands.build import build


SHORT_HELP = "build and run tests"
LONG_HELP = """
Usage: reseasdk test
"""


def test(args):
    cs = build(args)
    defaults = cs['defaults']

    info('ReseaSDK: build finished successfully, starting the executable...')
    env = dict_to_strdict(defaults)
    cmd = [defaults['HAL_RUN'], defaults['BUILD_DIR'] + '/application']
    run_emulator(cmd, save_log='build/{}/boot.log'.format(args['target']),
                 env=env, test=True)


class FSEventHandler(watchdog.events.FileSystemEventHandler):

    def __init__(self):
        super().__init__()
        self.lock = threading.Lock()
        self.last_file = ''
        self.last_hash = ''

    def on_any_event(self, event):
        def run_test():
            argv = list(filter(lambda x: x != '-P', sys.argv))
            subprocess.Popen(argv).wait()
            info('ReseaSDK: watching changes (Ctrl-D to quit)')

        def is_target_ext(path):
            exts = [
                'c', 'h', 'S',
                'yml',
                'ld',
                'cfg',
                'sh', 'py'
            ]
            return os.path.splitext(path)[1].lstrip('.') in exts

        with self.lock:
            # ignore directory changes
            if not os.path.isdir(event.src_path):
                # XXX: exclude files in the build directory
                if '/build/test/' in event.src_path:
                    return

                try:
                    with open(event.src_path, 'rb') as f:
                        hash = hashlib.sha1(f.read()).hexdigest()
                except:
                    # the file may be deleted (e.g. Emacs backup files)
                    return

                path = os.path.basename(event.src_path)

                if (is_target_ext(path) and
                   (self.last_hash != hash or self.last_file != path)):
                    self.last_hash = hash
                    self.last_file = path
                    info('ReseaSDK: detected changes in files'
                         '-- starting a test')
                    threading.Thread(target=run_test, daemon=True).start()


def main(args_):
    parser = argparse.ArgumentParser(prog='reseasdk test',
                                     description='test an executable')
    parser.add_argument('-r', action='store_true',
                        help='rebuild the executable')
    parser.add_argument('-P', action='store_true',
                        help='run tests automatically when'
                             'it detect a changes in files')
    parser.add_argument('--target', default='test', help='the build target')
    parser.add_argument('--dump', action='store_true',
                        help='dump generated files for debugging')
    parser.add_argument('--no-prettify', action='store_true',
                        help="don't prettify output")
    parser.add_argument('--all-in-one', action='store_true',
                        help='embed all required applications so that'
                             'discovery() always succeed')
    parser.add_argument('--single-app', action='store_true',
                        help='no threading')
    parser.add_argument('defaults', nargs='*',
                        help='default variables (FOO=bar)')

    args = vars(parser.parse_args(args_))
    args['defaults'].append('TEST=yes')

    if args['P']:
        # look for .git
        cwd = os.getcwd()
        while os.getcwd() != '/':
            if os.path.exists('.git'):
                path = os.getcwd()
            os.chdir('..')
        if os.getcwd() != '/':
            path = os.getcwd()
        os.chdir(cwd)

        info('ReseaSDK: watching changes in {} (Ctrl-D to quit)'.format(path))
        observer = watchdog.observers.Observer()
        observer.schedule(FSEventHandler(), path, recursive=True)
        observer.daemon = True
        observer.start()
        try:
            sys.stdin.read(1)
        except KeyboardInterrupt:
            pass

        info('ReseaSDK: stopping')
        observer.stop()
        observer.join()
    else:
        test(args)



import os
import sys
import atexit
import datetime
import subprocess
from termcolor import cprint, colored
from reseasdk.helpers import info, error


def lprint(s):
    """Prints a log message beautifully."""

    # try to get the column length of terminal
    try:
        columns = int(subprocess.check_outptu('stty size', 'r').split()[1])
    except:
        columns = 80

    try:
        # log message format is '[package_name] TYPE: a message from the package'
        if not s.startswith("["):
            raise ValueError
        package_name, rest = s.split(' ', 1)
        package_name = package_name.lstrip('[').rstrip(']')
        type_, rest = rest.split(':', 1)
        type_ = type_.rstrip(':')
        body = rest.strip()
    except ValueError:
        # invalid log message format: just print it
        print(s)
    else:
        # valid log message
        if type_ == 'TEST':
            # test results
            if body in ['start', 'end']:
                # 
                return
            elif body.startswith("<pass>"):
                type_ = 'PASS'
            else:
                type_ = 'FAIL'

        # pick a color                
        t = {
            'INFO':  (' I ', 'white', 'on_blue'),
            'BUG':   (' B ', 'white', 'on_red'),
            'PANIC': (' P ', 'white', 'on_red'),
            'WARN':  (' W ', 'white', 'on_yellow'),
            'DEBUG': (' D ', 'grey',  'on_white'),
            'PASS':  (' T ', 'white', 'on_green'),
            'FAIL':  (' T ', 'white', 'on_red'),
        }.get(type_, (' ',))

        pad_str = ' ' * (12 - len(package_name))
        package_name_str = colored(package_name, attrs=['bold'])
        type_str = colored(*t)
        print('{}{} {} '.format(pad_str, package_name_str, type_str), end='')

        column = 17 # the length of a string printed above
        for word in body.split(' '):
            if column + len(word) + 1 < columns:
                print(word + ' ', end='')
                column += len(word) + 1
            else:
                # line wrap
                print('\n{}{} '.format(' ' * 17, word, ' '), end='')
                column = 17 + len(word) + 1
        print('')

        if type_ == 'PANIC':
            # kernel panic: raise SystemExit to abort the emulator
            raise SystemExit


def try_parse(l):
    """
    Tries to TEST log messages. Returns 'pass' or 'fail' if `l` is a TEST log
    message or '' (an empty string) if it failed to parse.
    """
    try:
        test = l.split('TEST: ')[1]
        result = test.split(' ', 2)[0].lstrip('<').rstrip('>')
    except IndexError:
        result = ''
    return result


def atexit_handler(p):
    info('ReseaSDK: terminating the executable...')
    try:
        if p.poll() is not None:
            p.kill()
    except ProcessLookupError:
        pass


def run_emulator(cmd, test=False, env=None, save_log=None):
    if save_log is None:
        save_log = '/dev/null'

    # prepend a header to the log file
    f = open(save_log, 'a')
    f.write('================ {} ================\n'.format(
            str(datetime.datetime.now())))

    # run test program speicified in HAL_RUN config
    envs = os.environ.copy()
    envs.update(env)
    p = subprocess.Popen(cmd, env=envs,
                         stderr=subprocess.PIPE,
                         stdout=subprocess.PIPE)

    atexit.register(atexit_handler, p)

    # parse log messages
    passed = 0
    failed = 0
    while True:
        l = p.stdout.readline().decode('utf-8').strip()

        if l == '':
            p.terminate()
            p.kill()
            if test:
                error("the test program finished without 'TEST: end'")
            sys.exit(0)

        result = try_parse(l)
        if result == 'end':
            f.write(l + '\n')
            lprint(l)
            if test:
                if failed == 0:
                    cprint('ReseaSDK: All {} tests passed'.format(passed),
                           'green')
                else:
                    cprint('ReseaSDK: {} tests failed'.format(failed),
                           'red')
            p.terminate()
            p.kill()
            sys.exit(0)
        elif result == 'pass':
            passed += 1
            f.write(l + '\n')
            lprint(l)
        elif result == 'fail':
            failed += 1
            f.write(l + '\n')
            lprint(l)
        else:
            f.write(l + '\n')
            try:
                lprint(l)
            except SystemExit:
                # kernel panic
                p.terminate()
                p.kill()
                sys.exit(0)



from reseasdk.version import __version__
from reseasdk.main import main

import reseasdk.helpers
import reseasdk.validators

command_list = [
    'build',
    'test',
    'clean',
    'log',
    'new',
    'version'
]



from setuptools import find_packages, setup

# XXX: read the ReseaSDK's version
exec(open("reseasdk/version.py").read())


setup_args = {
    'name': 'ReseaSDK',
    'version': __version__,
    'url': 'http://resea.net',
    'author': 'Seiya Nuta',
    'author_email': 'nuta@seiya.me',
    'description': 'A software development kit for Resea',
    'license': 'Public Domain',
    'packages': ['reseasdk', 'reseasdk.commands'],
    'scripts': ['bin/reseasdk'],
    'install_requires': open('requirements.txt').read().strip().splitlines(),
    'classifiers': [
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: Public Domain',
        'Programming Language :: Python :: 3'
    ],
}


if __name__ == '__main__':
    setup(**setup_args)



from reseasdk.helpers import import_module
from reseasdk import command_list


def main(args):
    try:
        cmd = args[0]
    except IndexError:
        cmd = ''

    if cmd == '':
        print("Usage: reseasdk command")
        print("")
        print("Commands:")

        for c in command_list:
            m = import_module("reseasdk.commands.{}".format(c))
            print("  {:<10}{}".format(c, m.SHORT_HELP))
    else:
        m = import_module("reseasdk.commands.{}".format(cmd))
        print(m.LONG_HELP.strip() + "\n")
